@@ -14,6 +14,7 @@ import errno
 import mmap
 
 from contextlib import contextmanager
+import signal
 from subprocess import (
     call,
     Popen,
@@ -41,9 +42,10 @@ from git.compat import (
 
 execute_kwargs = ('istream', 'with_keep_cwd', 'with_extended_output',
                   'with_exceptions', 'as_process', 'stdout_as_string',
-                  'output_stream')
+                  'output_stream', 'with_stdout', 'kill_after_timeout')
 
 log = logging.getLogger('git.cmd')
+log.addHandler(logging.NullHandler())
 
 __all__ = ('Git', )
 
@@ -74,7 +76,10 @@ def handle_process_output(process, stdout_handler, stderr_handler, finalizer):
     :param stdout_handler: f(stdout_line_string), or None
     :param stderr_hanlder: f(stderr_line_string), or None
     :param finalizer: f(proc) - wait for proc to finish"""
-    def parse_lines_from_buffer(fno, buf):
+    fdmap = {process.stdout.fileno(): (stdout_handler, [b'']),
+             process.stderr.fileno(): (stderr_handler, [b''])}
+
+    def _parse_lines_from_buffer(buf):
         line = b''
         bi = 0
         lb = len(buf)
@@ -91,19 +96,19 @@ def handle_process_output(process, stdout_handler, stderr_handler, finalizer):
         # END while file is not done reading
     # end
 
-    def read_lines_from_fno(fno, last_buf_list):
+    def _read_lines_from_fno(fno, last_buf_list):
         buf = os.read(fno, mmap.PAGESIZE)
         buf = last_buf_list[0] + buf
 
         bi = 0
-        for bi, line in parse_lines_from_buffer(fno, buf):
+        for bi, line in _parse_lines_from_buffer(buf):
             yield line
         # for each line to parse from the buffer
 
         # keep remainder
         last_buf_list[0] = buf[bi:]
 
-    def dispatch_single_line(line, handler):
+    def _dispatch_single_line(line, handler):
         line = line.decode(defenc)
         if line and handler:
             try:
@@ -115,32 +120,34 @@ def handle_process_output(process, stdout_handler, stderr_handler, finalizer):
         # end dispatch helper
     # end single line helper
 
-    def dispatch_lines(fno, handler, buf_list):
+    def _dispatch_lines(fno, handler, buf_list):
         lc = 0
-        for line in read_lines_from_fno(fno, buf_list):
-            dispatch_single_line(line, handler)
+        for line in _read_lines_from_fno(fno, buf_list):
+            _dispatch_single_line(line, handler)
             lc += 1
         # for each line
         return lc
     # end
 
-    def deplete_buffer(fno, handler, buf_list, wg=None):
+    def _deplete_buffer(fno, handler, buf_list, wg=None):
+        lc = 0
         while True:
-            line_count = dispatch_lines(fno, handler, buf_list)
+            line_count = _dispatch_lines(fno, handler, buf_list)
+            lc += line_count
             if line_count == 0:
                 break
         # end deplete buffer
 
         if buf_list[0]:
-            dispatch_single_line(buf_list[0], handler)
+            _dispatch_single_line(buf_list[0], handler)
+            lc += 1
         # end
 
         if wg:
             wg.done()
-    # end
 
-    fdmap = {process.stdout.fileno(): (stdout_handler, [b'']),
-             process.stderr.fileno(): (stderr_handler, [b''])}
+        return lc
+    # end
 
     if hasattr(select, 'poll'):
         # poll is preferred, as select is limited to file handles up to 1024 ... . This could otherwise be
@@ -168,7 +175,7 @@ def handle_process_output(process, stdout_handler, stderr_handler, finalizer):
                 if result & CLOSED:
                     closed_streams.add(fd)
                 else:
-                    dispatch_lines(fd, *fdmap[fd])
+                    _dispatch_lines(fd, *fdmap[fd])
                 # end handle closed stream
             # end for each poll-result tuple
 
@@ -179,7 +186,7 @@ def handle_process_output(process, stdout_handler, stderr_handler, finalizer):
 
         # Depelete all remaining buffers
         for fno, (handler, buf_list) in fdmap.items():
-            deplete_buffer(fno, handler, buf_list)
+            _deplete_buffer(fno, handler, buf_list)
         # end for each file handle
 
         for fno in fdmap.keys():
@@ -193,7 +200,7 @@ def handle_process_output(process, stdout_handler, stderr_handler, finalizer):
         wg = WaitGroup()
         for fno, (handler, buf_list) in fdmap.items():
             wg.add(1)
-            t = threading.Thread(target=lambda: deplete_buffer(fno, handler, buf_list, wg))
+            t = threading.Thread(target=lambda: _deplete_buffer(fno, handler, buf_list, wg))
             t.start()
         # end
         # NOTE: Just joining threads can possibly fail as there is a gap between .start() and when it's
@@ -272,8 +279,10 @@ class Git(LazyMixin):
             self.proc = None
             if proc.stdin:
                 proc.stdin.close()
-            proc.stdout.close()
-            proc.stderr.close()
+            if proc.stdout:
+                proc.stdout.close()
+            if proc.stderr:
+                proc.stderr.close()
 
             # did the process finish already so we have a return code ?
             if proc.poll() is not None:
@@ -300,13 +309,23 @@ class Git(LazyMixin):
         def __getattr__(self, attr):
             return getattr(self.proc, attr)
 
-        def wait(self):
+        def wait(self, stderr=None):
             """Wait for the process and return its status code.
 
+            :param stderr: Previously read value of stderr, in case stderr is already closed.
+            :warn: may deadlock if output or error pipes are used and not handled separately.
             :raise GitCommandError: if the return status is not 0"""
             status = self.proc.wait()
+
+            def read_all_from_possibly_closed_stream(stream):
+                try:
+                    return stream.read()
+                except ValueError:
+                    return stderr or ''
+
             if status != 0:
-                raise GitCommandError(self.args, status, self.proc.stderr.read())
+                errstr = read_all_from_possibly_closed_stream(self.proc.stderr)
+                raise GitCommandError(self.args, status, errstr)
             # END status handling
             return status
     # END auto interrupt
@@ -336,7 +355,7 @@ class Git(LazyMixin):
         def read(self, size=-1):
             bytes_left = self._size - self._nbr
             if bytes_left == 0:
-                return ''
+                return b''
             if size > -1:
                 # assure we don't try to read past our limit
                 size = min(bytes_left, size)
@@ -355,7 +374,7 @@ class Git(LazyMixin):
 
         def readline(self, size=-1):
             if self._nbr == self._size:
-                return ''
+                return b''
 
             # clamp size to lowest allowed value
             bytes_left = self._size - self._nbr
@@ -469,6 +488,8 @@ class Git(LazyMixin):
                 as_process=False,
                 output_stream=None,
                 stdout_as_string=True,
+                kill_after_timeout=None,
+                with_stdout=True,
                 **subprocess_kwargs
                 ):
         """Handles executing the command on the shell and consumes and returns
@@ -522,6 +543,18 @@ class Git(LazyMixin):
             some of the valid kwargs are already set by this method, the ones you
             specify may not be the same ones.
 
+        :param with_stdout: If True, default True, we open stdout on the created process
+
+        :param kill_after_timeout:
+            To specify a timeout in seconds for the git command, after which the process
+            should be killed. This will have no effect if as_process is set to True. It is
+            set to None by default and will let the process run until the timeout is
+            explicitly specified. This feature is not supported on Windows. It's also worth
+            noting that kill_after_timeout uses SIGKILL, which can have negative side
+            effects on a repository. For example, stale locks in case of git gc could
+            render the repository incapable of accepting changes until the lock is manually
+            removed.
+
         :return:
             * str(output) if extended_output = False (Default)
             * tuple(int(status), str(stdout), str(stderr)) if extended_output = True
@@ -530,7 +563,7 @@ class Git(LazyMixin):
             * output_stream if extended_output = False
             * tuple(int(status), output_stream, str(stderr)) if extended_output = True
 
-            Note git is executed with LC_MESSAGES="C" to ensure consitent
+            Note git is executed with LC_MESSAGES="C" to ensure consistent
             output regardless of system language.
 
         :raise GitCommandError:
@@ -549,11 +582,18 @@ class Git(LazyMixin):
 
         # Start the process
         env = os.environ.copy()
-        env["LC_MESSAGES"] = "C"
+        # Attempt to force all output to plain ascii english, which is what some parsing code
+        # may expect.
+        # According to stackoverflow (http://goo.gl/l74GC8), we are setting LANGUAGE as well
+        # just to be sure.
+        env["LANGUAGE"] = "C"
+        env["LC_ALL"] = "C"
         env.update(self._environment)
 
         if sys.platform == 'win32':
             cmd_not_found_exception = WindowsError
+            if kill_after_timeout:
+                raise GitCommandError('"kill_after_timeout" feature is not supported on Windows.')
         else:
             if sys.version_info[0] > 2:
                 cmd_not_found_exception = FileNotFoundError  # NOQA # this is defined, but flake8 doesn't know
@@ -565,9 +605,10 @@ class Git(LazyMixin):
             proc = Popen(command,
                          env=env,
                          cwd=cwd,
+                         bufsize=-1,
                          stdin=istream,
                          stderr=PIPE,
-                         stdout=PIPE,
+                         stdout=with_stdout and PIPE or None,
                          shell=self.USE_SHELL,
                          close_fds=(os.name == 'posix'),  # unsupported on windows
                          **subprocess_kwargs
@@ -578,13 +619,50 @@ class Git(LazyMixin):
         if as_process:
             return self.AutoInterrupt(proc, command)
 
+        def _kill_process(pid):
+            """ Callback method to kill a process. """
+            p = Popen(['ps', '--ppid', str(pid)], stdout=PIPE)
+            child_pids = []
+            for line in p.stdout:
+                if len(line.split()) > 0:
+                    local_pid = (line.split())[0]
+                    if local_pid.isdigit():
+                        child_pids.append(int(local_pid))
+            try:
+                # Windows does not have SIGKILL, so use SIGTERM instead
+                sig = getattr(signal, 'SIGKILL', signal.SIGTERM)
+                os.kill(pid, sig)
+                for child_pid in child_pids:
+                    try:
+                        os.kill(child_pid, sig)
+                    except OSError:
+                        pass
+                kill_check.set()    # tell the main routine that the process was killed
+            except OSError:
+                # It is possible that the process gets completed in the duration after timeout
+                # happens and before we try to kill the process.
+                pass
+            return
+        # end
+
+        if kill_after_timeout:
+            kill_check = threading.Event()
+            watchdog = threading.Timer(kill_after_timeout, _kill_process, args=(proc.pid, ))
+
         # Wait for the process to return
         status = 0
         stdout_value = b''
         stderr_value = b''
         try:
             if output_stream is None:
+                if kill_after_timeout:
+                    watchdog.start()
                 stdout_value, stderr_value = proc.communicate()
+                if kill_after_timeout:
+                    watchdog.cancel()
+                    if kill_check.isSet():
+                        stderr_value = 'Timeout: the command "%s" did not complete in %d ' \
+                                       'secs.' % (" ".join(command), kill_after_timeout)
                 # strip trailing "\n"
                 if stdout_value.endswith(b"\n"):
                     stdout_value = stdout_value[:-1]
@@ -686,23 +764,31 @@ class Git(LazyMixin):
         finally:
             self.update_environment(**old_env)
 
-    def transform_kwargs(self, split_single_char_options=False, **kwargs):
+    def transform_kwarg(self, name, value, split_single_char_options):
+        if len(name) == 1:
+            if value is True:
+                return ["-%s" % name]
+            elif type(value) is not bool:
+                if split_single_char_options:
+                    return ["-%s" % name, "%s" % value]
+                else:
+                    return ["-%s%s" % (name, value)]
+        else:
+            if value is True:
+                return ["--%s" % dashify(name)]
+            elif type(value) is not bool:
+                return ["--%s=%s" % (dashify(name), value)]
+        return []
+
+    def transform_kwargs(self, split_single_char_options=True, **kwargs):
         """Transforms Python style kwargs into git command line options."""
         args = list()
         for k, v in kwargs.items():
-            if len(k) == 1:
-                if v is True:
-                    args.append("-%s" % k)
-                elif type(v) is not bool:
-                    if split_single_char_options:
-                        args.extend(["-%s" % k, "%s" % v])
-                    else:
-                        args.append("-%s%s" % (k, v))
+            if isinstance(v, (list, tuple)):
+                for value in v:
+                    args += self.transform_kwarg(k, value, split_single_char_options)
             else:
-                if v is True:
-                    args.append("--%s" % dashify(k))
-                elif type(v) is not bool:
-                    args.append("--%s=%s" % (dashify(k), v))
+                args += self.transform_kwarg(k, v, split_single_char_options)
         return args
 
     @classmethod
