@@ -186,7 +186,7 @@ class FetchInfo(object):
         FAST_FORWARD, ERROR = [1 << x for x in range(8)]
 
     #                             %c    %-*s %-*s             -> %s       (%s)
-    re_fetch_result = re.compile("^\s*(.) (\[?[\w\s\.]+\]?)\s+(.+) -> ([/\w_\+\.\-#]+)(    \(.*\)?$)?")
+    re_fetch_result = re.compile("^\s*(.) (\[?[\w\s\.$]+\]?)\s+(.+) -> ([/\w_\+\.\-$#]+)(    \(.*\)?$)?")
 
     _flag_map = {'!': ERROR,
                  '+': FORCED_UPDATE,
@@ -607,11 +607,12 @@ class Remote(LazyMixin, Iterable):
     def _assert_refspec(self):
         """Turns out we can't deal with remotes if the refspec is missing"""
         config = self.config_reader
+        unset = 'placeholder'
         try:
-            if config.get_value('fetch', default=type) is type:
+            if config.get_value('fetch', default=unset) is unset:
                 msg = "Remote '%s' has no refspec set.\n"
                 msg += "You can set it as follows:"
-                msg += " 'git config --add \"remote.%s.fetch +refs/heads/*:refs/heads/*\"'." % self.name
+                msg += " 'git config --add \"remote.%s.fetch +refs/heads/*:refs/heads/*\"'." % (self.name, self.name)
                 raise AssertionError(msg)
         finally:
             config.release()
@@ -643,7 +644,9 @@ class Remote(LazyMixin, Iterable):
         :note:
             As fetch does not provide progress information to non-ttys, we cannot make
             it available here unfortunately as in the 'push' method."""
-        self._assert_refspec()
+        if refspec is None:
+            # No argument refspec, then ensure the repo's config has a fetch refspec.
+            self._assert_refspec()
         kwargs = add_progress(kwargs, self.repo.git, progress)
         if isinstance(refspec, list):
             args = refspec
@@ -665,7 +668,9 @@ class Remote(LazyMixin, Iterable):
         :param progress: see 'push' method
         :param kwargs: Additional arguments to be passed to git-pull
         :return: Please see 'fetch' method """
-        self._assert_refspec()
+        if refspec is None:
+            # No argument refspec, then ensure the repo's config has a fetch refspec.
+            self._assert_refspec()
         kwargs = add_progress(kwargs, self.repo.git, progress)
         proc = self.repo.git.pull(self, refspec, with_stdout=False, as_process=True, v=True, **kwargs)
         res = self._get_fetch_info_from_stderr(proc, progress or RemoteProgress())
